/**************************************************************
 * FONTFILT.C	Version 2.0                  (Turbo C 2.0)
 *
 * Copyright (c) 1988  Jim Bumgardner
 *
 * Revision History
 *	   Version 2.0   11/88		 Jim Bumgardner
 *		- Combined 11 different MAKE* programs into FONTFILT.C
 *		  Everything now resides in one sourcefile.
 *
 *         Version 1.2   10/88           Jim Bumgardner
 *              - Using regular bitmapped translation (as opposed to
 *		  bit->byte of Version 1.0.
 *		- Put routines in a library.
 *		- Sped up outlining and smearing routines.
 *		- Added HALF, FUZZ, INVERT, FIXED, BOLD and BOX
 *
 *	   Version 1.1   01/13/88        Pete Wilson
 *         ("FFILTERS.H" from XFNTFILT.ARC)
 *		- 60 pt fonts supported by compiling LARGE model
 *		- Added ANSI compatible code and MSC support
 *
 *	   Version 1.0   10/87           Jim Bumgardner
 *         ("FFILTERS.H" from FONTFILT.ARC)
 ***************************************************************/

#include <stdio.h>
#include <ctype.h>

#define NO_EFFECTS	11
#define MAX_COMBINE	10
#define TRUE		1
#define FALSE		0

void boldfont(),boxfont(),doubfont(),dracfont(),fixfont(),fuzzfont(),
     halffont(),invfont(),outfont(),shadfont();

void boldchar(),boxchar(),doubchar(),dracchar(),fixchar(),fuzzchar(),
     halfchar(),invchar(),oshdchar(),outchar(),shadchar();

struct effects
{
	char *name;
	int default1,default2;
	void (*fontfunc)();
	void (*charfunc)();
} effects[NO_EFFECTS] = {
	"BOLD hsmear vsmear",			1,0,boldfont,boldchar,
	"BOX clearance",			2,0,boxfont,boxchar,
	"DOUBLE",				0,0,doubfont,doubchar,
	"DRACULA seed maxdrip",			0,14,dracfont,dracchar,
	"FIXED",				0,0,fixfont,fixchar,
	"FUZZY",				0,0,fuzzfont,fuzzchar,
	"HALF",					0,0,halffont,halfchar,
	"INVERT hflush vflush",			0,0,invfont,invchar,
	"OPENSHADOW smearamt thickness",	10,1,shadfont,oshdchar,
	"OUTLINE thickness",			1,0,outfont,outchar,
	"SHADOW smearamt thickness",		10,1,shadfont,shadchar};

int noreqs;

struct request {
	int command;
	int arg1,arg2;
} reqs[MAX_COMBINE];

syntax()
{
	int i;
	fputs(
		"FONTFILT Version 2.0\t\t\tCopyright (c) 1988 Jim Bumgardner\n\n"
		"Syntax:\n"
		"\tFONTFILT inputfont outputfont [effect [effect...]]\n\n"
		"Effects:\n",stderr);
	for (i =0; i < NO_EFFECTS; ++i)
		fprintf(stderr,"\t%s\n",effects[i].name);
	exit(0);
}


main(int argc, char *argv[])
{
	int i,j;
	if (argc < 3)	syntax();
	for (i = 3; i < argc; ++i)
	{
		for (j = 0; j < NO_EFFECTS; ++j)
			if (strnicmp(argv[i],effects[j].name,strlen(argv[i])) == 0)
				break;
		if (j >= NO_EFFECTS)
			erxit("%s? Enter FONTFILT ? for help\n",argv[i]);
		if (noreqs == MAX_COMBINE)
			erxit("Maximum of %d Effects",MAX_COMBINE);
		reqs[noreqs].command = j;
		reqs[noreqs].arg1 = effects[j].default1;
		reqs[noreqs].arg2 = effects[j].default2;
		if (i < argc-1 && isdigit(argv[i+1][0]))
		{
			++i;
			reqs[noreqs].arg1 = atoi(argv[i]);
			if (i < argc-1 && isdigit(argv[i+1][0]))
			{
				++i;
				reqs[noreqs].arg2 = atoi(argv[i]);
			}
		}
		++noreqs;
	}
	convert_font(argv[1],argv[2]);
}

/*****************************************************************
 * Generalized Font Filtering Routines
 *
 * These functions call individual font/char header modification
 * functions for each effect.
 ******************************************************************/

#define YSIZE_LIMIT	256
#define XSIZE_LIMIT	256
#define MAXBUFSIZE	(YSIZE_LIMIT*(XSIZE_LIMIT>>3))
#define ESC		27
#define TWO_CHARS(c1, c2) (((c2) << 8) + (c1))

extern int nobufs;	/* Default size is 4 bitmap buffers */

/* Basic bit-access Macros */
#define isset(f,y,x)	(pmap[f][(y)*(XSIZE_LIMIT>>3)+((x)>>3)] & (0x80 >> ((x) & 7)))
#define doset(f,y,x)	(pmap[f][(y)*(XSIZE_LIMIT>>3)+((x)>>3)] |= 0x80 >> ((x) & 7))
#define unset(f,y,x)	(pmap[f][(y)*(XSIZE_LIMIT>>3)+((x)>>3)] &= ~(0x80 >> ((x) & 7)))
#define xorset(f,x,y)	(pmap[f][(y)*(XSIZE_LIMIT>>3)+((x)>>3)] ^= 0x80 >> ((x) & 7))

/* Byte-access macros */
#define bits2bytes(x)	(((x-1)>>3)+1)
#define pmapbyte(f,y,x)	pmap[f][(y)*(XSIZE_LIMIT>>3)+x]
#define pmaprow(f,y)	&pmap[f][(y)*(XSIZE_LIMIT>>3)]
#define clearchar(f)	memset(pmap[f],0,MAXBUFSIZE)
#define copychar(t,f)	memcpy(pmap[t],pmap[f],MAXBUFSIZE)

struct	font_header {
	int fil26;
	char nul1,font_type;
	int nul2,baseline,width,height;
	char orient,spacing;
	int symset,pitch,points,nul3;
	char nul4,style,weight,type_face;
} font_header;

struct char_header {
	char fil4,nul1,fil14,cil1;
	char orient,nul2;
	int left_offset,top_offset,width,height,delta_x;
} char_header;

int nobufs = 4;			/* Change this if you need more bitmaps */
unsigned char **pmap;
unsigned char *ibuf,*bitmap;

FILE *ifile,*ofile;

convert_font(char *ifname,char *ofname)
{
	int total_bytes,horiz_bytes;
	int i;
	void *malloc();
	
	/* Allocate Memory */
	if ((ibuf = malloc(MAXBUFSIZE+sizeof(char_header))) == NULL ||
	    (pmap = malloc(sizeof(char *) * nobufs)) == NULL)
		erxit("Out of Memory");
		
	for (i = 0; i < nobufs; ++i)
		if ((pmap[i] = malloc(MAXBUFSIZE)) == NULL)
			erxit("Out of Memory");

	bitmap = &ibuf[sizeof(char_header)];

	if ((ifile = fopen(ifname,"rb")) == NULL)
		erxit("Can't open Input File %s\n");
	if ((ofile = fopen(ofname,"wb")) == NULL)
		erxit("Can't open Output File %s\n");

	if ((total_bytes = get_font_header()) == EOF)
		erxit("Invalid Font File");

	for (i = 0; i < noreqs; ++i)
		(*effects[reqs[i].command].fontfunc)(reqs[i].arg1,reqs[i].arg2);

	write_font_header(total_bytes);

	printf("Processing  ");
	while ((total_bytes = get_char_header()) != EOF)
	{
		memcpy(&char_header,ibuf,sizeof(char_header));
		swapcharheader();

		horiz_bytes = bits2bytes(char_header.width);
		translate_bitmap(char_header.height,horiz_bytes);
		for (i = 0; i < noreqs; ++i)
		{
			clearchar(1);
			(*effects[reqs[i].command].charfunc)();
			if (i < noreqs-1)
				copychar(0,1);
		}			
		horiz_bytes = bits2bytes(char_header.width);
		translate_pmap(char_header.height,horiz_bytes);

		total_bytes = sizeof(char_header)+ horiz_bytes * char_header.height;
		swapcharheader();
		memcpy(ibuf,&char_header,sizeof(char_header));
		fprintf(ofile,"\033(s%dW",total_bytes);
		fwrite(ibuf,total_bytes,1,ofile);
	}
	fclose(ifile);
	fclose(ofile);
}

translate_bitmap(int h,int w)
{
	int y;
	char *p;

	clearchar(0);
	p = bitmap;
	for (y = 0; y < h; ++y)
	{
		memcpy(pmaprow(0,y),p,w);
		p += w;
	}
}

translate_pmap(int h,int w)
{
	int y;
	char *p;
	p = bitmap;
	for (y = 0; y < h; ++y)
	{
		memcpy(p,pmaprow(1,y),w);
		p += w;
	}
}

get_font_header()
{
	unsigned int tmp;
	int c;

	do {
		while ((c = getc(ifile)) != EOF && c != ESC)
			;
		if (c == EOF)
			erxit("Invalid Font File");
		tmp = getw(ifile);
	} while (tmp != TWO_CHARS(')','s'));

	tmp = get_number();
	fread(ibuf,tmp,1,ifile);
	memcpy(&font_header,ibuf,sizeof(font_header));
	swapfontheader();	/* Make Header Integers Legible */
	if (font_header.width > XSIZE_LIMIT ||
	    font_header.height > YSIZE_LIMIT)
	{
		fprintf(stderr,"Characters are too big\n");
		return(EOF);
	}
	return(tmp);
}

write_font_header(int t)
{
	fprintf(ofile,"\033)s%dW",t);
	swapfontheader();
	memcpy(ibuf,&font_header,sizeof(font_header));
	fwrite(ibuf,t,1,ofile);
	swapfontheader();	/* Make Header Integers Legible Again */
}

get_char_header()
{
	int tmp;
	int c,cno;
	while ((c = getc(ifile)) != EOF && c != ESC)
		;
	if (c == EOF)		return(EOF);
	tmp = getw(ifile);

	if (tmp != TWO_CHARS('*','c'))	return(EOF);
	cno = get_number();

	printf("\b%c",cno);
	fprintf(ofile,"\033*c%dE",cno);

	c = getc(ifile);		/* skip escape */
	tmp = getw(ifile);
	if (tmp != TWO_CHARS('(','s'))	return(EOF);

	tmp = get_number();

	fread(ibuf,tmp,1,ifile);
	return(tmp);
}

swapfontheader()
{
	font_header.baseline = iswap(font_header.baseline);
	font_header.width = iswap(font_header.width);
	font_header.height = iswap(font_header.height);
	font_header.pitch = iswap(font_header.pitch);
	font_header.points = iswap(font_header.points);
}

swapcharheader()
{
	char_header.width = iswap(char_header.width);
	char_header.height = iswap(char_header.height);
	char_header.delta_x = iswap(char_header.delta_x);
	char_header.left_offset = iswap(char_header.left_offset);
	char_header.top_offset = iswap(char_header.top_offset);
}

iswap(unsigned int n)
{
	return((n >> 8) | (n << 8));
}

get_number()
{
	char *p;
	int c;
	p = ibuf;
	while ((c = getc(ifile)) != EOF && isdigit(c))
		*p++ = c;
	*p = 0;
	return(atoi(ibuf));
}

erxit(char *tmp,...)
{
	va_list arg_ptr;
	va_start(arg_ptr,tmp);
	vfprintf(stderr,tmp,arg_ptr);
	va_end(arg_ptr);
	exit(0);
}

/****************************************************************
 * Useful Generalized Filtering Routines
 ****************************************************************/

int rh,rw;

outline(int to, int from, int h, int w, int thickness,int quadrants)
{
	int i;
	int q;

	rh = h;
	rw = bits2bytes(w);

	for (q = 0; q < 4; ++q)
	{
		if (quadrants & (1 << q))
		{
			copychar(2,from);
			for (i = 0; i < thickness; ++i)
			{
				switch (q) {
				case 0:	shift_right(2+!(i&1),2+(i&1)); break;
				case 1: shift_down(2+!(i&1),2+(i&1));	break;
				case 2:	shift_left(2+!(i&1),2+(i&1));	break;
				case 3: shift_up(2+!(i&1),2+(i&1));	break;
				}
			}
			xormask(from,2+(i&1),to);
		}
	}
}

xormask(int fr1,int fr2,int to)
{
	register int y,x;

	for (y = 0; y < rh; ++y)
	{
		for (x = 0; x < rw; ++x)
		{
			pmapbyte(to,y,x) |= (pmapbyte(fr1,y,x) ^ 
					     pmapbyte(fr2,y,x)) &
					    pmapbyte(fr1,y,x);
		}
	}
}

shift_right(int to,int fr)
{
	register int y,x;
	for (y = 0; y < rh; ++y)
	{
		for (x = 0; x < rw; ++x)
		{
			pmapbyte(to,y,x) = pmapbyte(fr,y,x) >> 1;
			if (x && (pmapbyte(fr,y,x-1) & 1))
				pmapbyte(to,y,x) |= 0x80;
		}
	}
}

shift_left(int to,int fr)
{
	register int y,x;
	for (y = 0; y < rh; ++y)
	{
		for (x = 0; x < rw; ++x)
		{
			pmapbyte(to,y,x) = pmapbyte(fr,y,x) << 1;
			if (x < rw-1 && (pmapbyte(fr,y,x+1) & 0x80))
				pmapbyte(to,y,x) |= 1;
		}
	}
}

shift_down(int to,int fr)
{
	register int y;
	memset(pmaprow(to,0),0,rw);
	for (y = 0; y < rh; ++y)
		memcpy(pmaprow(to,y+1),pmaprow(fr,y),rw);
}

shift_up(int to,int fr)
{
	register int y;
	memset(pmaprow(to,rh-1),0,rw);
	for (y = 1; y < rh; ++y)
		memcpy(pmaprow(to,y-1),pmaprow(fr,y),rw);
}

/**********************
 * Smear
 *
 **********************/

smear(int to,int from,int h,int w,int smearamt) /* Smear down-right */
{
	register int i,y,x;

	rh = h;
	rw = bits2bytes(w);

	copychar(to,from);
	for (i = 1; i <= smearamt; ++i)
	{
		for (y = rh+i-1; y > 0; --y)
		{
			for (x = rw+i-1; x >= 0; --x)
			{
				pmapbyte(to,y,x) |= pmapbyte(to,y-1,x) >> 1;
				if (x && (pmapbyte(to,y-1,x-1) & 1))
					pmapbyte(to,y,x) |= 0x80;
			}
		}
	}
}

maskout(int to,int from,int h, int w)
{
	register int y,x;

	rh = h;
	rw = bits2bytes(w);

	for (y = 0; y < rh; ++y)
		for (x = 0; x < rw; ++x)
			pmapbyte(to,y,x) &= ~pmapbyte(from,y,x);
}

/********
 * BOLD *
 ********/

int hsmear=1,vsmear=0;

void boldfont(int hsm,int vsm)
{
	hsmear = hsm;
	vsmear = vsm;
	font_header.width += hsmear;
	font_header.height += vsmear;
	if (font_header.weight < 7)	font_header.weight++;
}

void boldchar()			/* Process pmap[0] --> pmap[1] */
{
	register int y,x;
	char i;
	int nxx,nyy;
	int h,w;

	h = char_header.height;
	w = char_header.width;
	nxx = hsmear;
	nyy = vsmear;
	while (char_header.width <= 256 && char_header.width+nxx > 256)
		--nxx;
	while (char_header.height <= 256 && char_header.height+nyy > 256)
		--nyy;

	char_header.width += nxx;
	char_header.height += nyy;
	char_header.delta_x += nxx;

	copychar(1,0);
	smeardown(1,h,w,nyy);
	smearright(1,h+nyy,w,nxx);
}

smeardown(int to,int h,int w,int smearamt) /* Smear down */
{
	register int i,y,x;
	int rh,rw;

	rh = h;
	rw = bits2bytes(w);

	for (i = 1; i <= smearamt; ++i)
	{
		for (y = rh+i-1; y > 0; --y)
		{
			for (x = rw-1; x >= 0; --x)
			{
				pmapbyte(to,y,x) |= pmapbyte(to,y-1,x);
			}
		}
	}
}

smearright(int to,int h,int w,int smearamt) /* Smear right */
{
	register int i,y,x;
	int rh,rw;

	rh = h;
	rw = bits2bytes(w);

	for (i = 1; i <= smearamt; ++i)
	{
		for (y = rh-1; y > 0; --y)
		{
			for (x = rw-1; x >= 0; --x)
			{
				pmapbyte(to,y,x) |= pmapbyte(to,y,x) >> 1;
				if (x && (pmapbyte(to,y,x-1) & 0x01))
					pmapbyte(to,y,x) |= 0x80;
			}
		}
	}
}

/*******
 * BOX *
 *******/

int clearance = 1;

void boxfont(int arg1)
{
	clearance = arg1;
	font_header.width += clearance*2;
	font_header.height += clearance*2;
	font_header.baseline += clearance;
}

void boxchar()
{
	int y,x,rw;
	int nw,nh;

	char_header.width += clearance*2;
	char_header.height += clearance*2;
	char_header.delta_x += clearance*2;
	char_header.top_offset +=clearance;
	char_header.left_offset += clearance;

	nw = char_header.width;
	nh = char_header.height;

	for (y = 0; y < nh; ++y)
	{
		for (x = 0; x < nw; ++x)
		{
			if (isset(0,y,x))
				doset(1,y+clearance,x+clearance);
			if (y == 0 || x == 0 || y == nh-1 || x == nw-1)
				doset(1,y,x);
		}
	}
}

/***********
 * DRACULA *
 ***********/

int maxdrip,seed;

void dracfont(int arg1, int arg2)
{
	seed = arg1;
	maxdrip = arg2;
	srand(seed);
	font_header.width++;
	font_header.height += maxdrip;
}

void dracchar()			/* Process pmap[0] --> pmap[1] */
{
	register int y,x,hd;
	register char i;
	int no_drips,drop;
	int w,h;

	w = char_header.width;
	h = char_header.height;

	no_drips = w/5;

	copychar(1,0);

	for (i = 0; i < no_drips; ++i)
	{
		x = rand() % w;

		if (x == 0)	continue;

		for (y = rand() % h; y > 0; --y)
			if (isset(0,y,x))
				break;

		if (y == 0 && isset(0,y,x) == 0)
			continue;

		while (isset(0,y,x))	++y;

		hd = 1 + rand() % (maxdrip - 1);

		drop = (hd > 7);
		if (drop)
			hd = 1 + rand() % 5;

		if (isset(0,y-1,x-1))
			doset(1,y,x-1);
		doset(1,y,x);
		if (isset(0,y-1,x+1))
			doset(1,y,x+1);
dodrop:
		for (;hd-- > 0; ++y)
		{
			if (hd == 1 || hd == 2)
			{
				doset(1,y,x-1);
				doset(1,y,x);
				doset(1,y,x+1);
			}
			else	doset(1,y,x);
		}
		if (drop) {
			++y;
			++y;
			hd = 3;
			drop = 0;
			goto dodrop;
		}
	}
	char_header.width++;
	char_header.height += maxdrip;
	char_header.delta_x++;
}

/*********
 * FIXED *
 *********/

void fixfont()
{
	font_header.spacing = 0;
	font_header.pitch = font_header.width << 2;
}

void fixchar()
{
	int h,w,y,x;
	int l;
	h = char_header.height;
	w = char_header.width;
	l = char_header.left_offset;
	if (l <= 0)
		copychar(1,0);
	else {
		for (y = 0; y < h; ++y)
			for (x = 0; x < w; ++x)
				if (isset(0,y,x))
					doset(1,y,x+l);
		char_header.width += l;
	}
	char_header.left_offset = 0;
	char_header.delta_x = 0;
}

/*********
 * FUZZY *
 *********/

void fuzzfont()
{
	font_header.width += 2;
	font_header.height += 2;
}

void fuzzchar()			/* Process pmap[0] --> pmap[1] */
{
	int w,h,x,y;
	w = char_header.width;
	h = char_header.height;
	for (y = 0; y < h; ++y)
	    for (x = 0; x < w; ++x)
		if (isset(0,y,x))
		{
			switch(rand() % 5)
			{
			case 0:	doset(1,y+1,x+1);	break;
			case 1:	doset(1,y,x);		break;
			case 2:	doset(1,y+2,x);		break;
			case 3:	doset(1,y,x);	 	break;
			case 4: doset(1,y,x+2);		break;
			}
		}
	char_header.width += 2;
	char_header.height += 2;
	char_header.left_offset--;
	char_header.top_offset--;
}

/********
 * HALF *
 ********/

void halffont()
{
	font_header.width >>= 1;
	font_header.height >>= 1;
	font_header.baseline >>= 1;
	font_header.pitch >>= 1;
	font_header.points >>= 1;
}

void halfchar()
{
	int y,x,h,w;
	h = char_header.height;
	w = char_header.width;

	for (y = 0; y < h; ++y)
		for (x = 0; x < w; ++x)
			if (isset(0,y,x))
				doset(1,y>>1,x>>1);
	char_header.width >>= 1;
	char_header.height >>= 1;
	char_header.delta_x >>= 1;
	char_header.top_offset >>= 1;
	char_header.left_offset >>= 1;
}

/**********
 * DOUBLE *
 **********/

void doubfont()
{
	font_header.width <<= 1;
	font_header.height <<= 1;
	font_header.baseline <<= 1;
	font_header.pitch <<= 1;
	font_header.points <<= 1;
}

void doubchar()
{
	int y,x,h,w;
	h = char_header.height;
	w = char_header.width;

	for (y = 0; y < h; ++y)
	    for (x = 0; x < w; ++x)
		if (isset(0,y,x))
		{
			doset(1,y<<1,x<<1);
			doset(1,(y<<1)+1,x<<1);
			doset(1,y<<1,(x<<1)+1);
			doset(1,(y<<1)+1,(x<<1)+1);
		}
	char_header.width <<= 1;
	char_header.height <<= 1;
	char_header.delta_x <<= 1;
	char_header.top_offset <<= 1;
	char_header.left_offset <<= 1;
}

/**********
 * INVERT *
 **********/

unsigned char hflush,vflush;

void invfont(int arg1, int arg2)
{
	hflush = arg1;
	vflush = arg2;
}

void invchar()
{
	int y,x,rw;
	int h,w,yo;
	
	if (hflush && char_header.width < (char_header.delta_x >> 2))
		char_header.width = (char_header.delta_x>>2);
		if (char_header.left_offset <= 0)
			char_header.width += -(char_header.left_offset) + 1;
	if (vflush && char_header.top_offset < font_header.baseline)
	{
		yo = font_header.baseline - char_header.top_offset;
		char_header.height += yo;
		char_header.top_offset += yo;
	}
	else
		yo = 0;	
	if (vflush)
		char_header.height = font_header.height;
	h = char_header.height;
	w = char_header.width;
	rw = bits2bytes(char_header.width);

	for (y = 0; y < h; ++y)
		for (x = 0; x < w; ++x)
			doset(1,y,x);

	for (y = 0; y < h; ++y)
		for (x = 0; x < rw; ++x)
			pmapbyte(1,y+yo,x) ^= pmapbyte(0,y,x);
}

/**************
 * SHADOW &
 * OPENSHADOW *
 **************/

int smearamt,thickness;

void shadfont(int arg1,int arg2)
{
	smearamt = arg1;
	thickness = arg2;
	font_header.width += smearamt;
	font_header.height += smearamt;
}

void shadchar()			/* Process pmap[0] --> pmap[1] */
{
	register int y,x;
	char i;
	int nsm;

	nsm = smearamt;
	while (char_header.width <= 256 && char_header.width+nsm > 256)
		--nsm;
	while (char_header.height <= 256 && char_header.height+nsm > 256)
		--nsm;
		
	char_header.width += nsm;
	char_header.height += nsm;
	char_header.delta_x += nsm;

	smear(1,0,char_header.height,char_header.width,nsm);
	maskout(1,0,char_header.height,char_header.width);
	outline(1,0,char_header.height,char_header.width,thickness,0x03);
}

void oshdchar()			/* Process pmap[0] --> pmap[1] */
{
	register int y,x;
	register char i;
	int nsm;

	nsm = smearamt;
	while (char_header.width <= 256 && char_header.width+nsm > 256)
		--nsm;
	while (char_header.height <= 256 && char_header.height+nsm > 256)
		--nsm;

	char_header.width += nsm;
	char_header.height += nsm;
	char_header.delta_x += nsm;

	smear(1,0,char_header.height,char_header.width,nsm);
	outline(0,1,char_header.height+smearamt,char_header.width+smearamt,thickness,0x0F);
	copychar(1,0);
}

/***********
 * OUTLINE *
 ***********/

int outlinethickness = 1;

void outfont(int arg1)
{
	outlinethickness = arg1;
}

void outchar()			/* Process pmap[0] --> pmap[1] */
{
	outline(1,0,char_header.height,char_header.width,outlinethickness,0x0F);
}
